from __future__ import unicode_literals


TAGTYPE_LIST = [
    'Artist',
    'ArtistSort',
    'Album',
    'AlbumArtist',
    'AlbumArtistSort',
    'Title',
    'Track',
    'Name',
    'Genre',
    'Date',
    'Composer',
    'Performer',
    'Disc',
    'MUSICBRAINZ_ARTISTID',
    'MUSICBRAINZ_ALBUMID',
    'MUSICBRAINZ_ALBUMARTISTID',
    'MUSICBRAINZ_TRACKID',
    'X-AlbumUri',
    'X-AlbumImage',
]
